/****************************************************************************
 *         Empfaenger Wetter-Funkbruecke als TWI-Slave
 *         ===========================================
 *                     Uwe Berger; 2013
 *
 * TWI-Slave-Lib:
 * --------------
 * -Martin Junghans
 * -http://www.jtronics.de/avr-projekte/library-i2c-twi-slave.html
 * 
 * RFM12-Lib:
 * ----------
 * -Benedikt K.
 * -http://www.mikrocontroller.net/topic/71682
 * 
 * 
 * Hardware
 * --------
 * -RFM12-Modul: Schaltung siehe http://www.mikrocontroller.net/topic/71682
 * -Hardware-TWI: siehe entsprechendes Atmel-Datenblatt zur MCU
 * 
 * Software
 * --------
 * -Empfang von Daten von einem entsprechenden Sender via Funk
 * -Ablage der empfangenen Datenbytes in gesendeter Reihenfolge in einem
 *  Puffer
 * -Pufferinhalt kann via TWI (I2C) abgefragt werden (Adressierung der
 *  entsprechenden Pufferadresse z.B. analog eine I2C-EEPROM...) 
 * 
 * 
 * ---------
 * Have fun! 
 *
 ***************************************************************************/
/*	
Anzeigeformatierung:

In C formatiert man floating-Point-Ausgaben mit der dafr geeigneten
Funktion printf(). Diese bentigt zwar viel Speicher und ist daher bei
einfacheren Aufgaben besser nicht zu verwenden, aber hier ist sie
praktisch.

Mit dem Formatspezifizierer "%n.mf"
  n - Feldbreite
  m - Przision (anzahl Nachkommastellen)

lsst sich so etwas hinbekommen.
Ist die Feldbreite lnger als die resultierende auszugebende Zahl, so
wird rechtsbndig mit fhrenden Leerzeichen ausgegeben.

Beispiel

  printf("%10.2f", x);

Durch Voransetzen eines '-' vor die Feldbreite wird linksbndig
ausgegeben und mit folgenden Leerzeichen aufgefllt.

  printf("%-10.2f", x);

Durch voransetzen einer 0 werden statt fhrender Leerzeichen fhrende
Nullen ausgegeben:

  printf("%010.2f", x);

(010 wird hier nicht als oktale Zahl interpretiert)

Diese Syntax kann auch bei nicht-floatingpoint-Zahlen und auch Strings
angewandt werden:

  printf("%08X", i);

  printf("%8d", i);

  printf("%13s", string);


Fr die Ausgabe auf ein LC-Display erscheint der Gebrauch von
snprintf() (printf in lngenbegrenzten Puffer) sinnvoll.	
*/

#include "config.h"	       //IMPORTANT: it contains F_CPU definition, so include it before delay.h
#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdint.h> 	    //definiert den Datentyp uint8_t

#include "lcd/HD44780.h"
#include "rfm12/rf12.h"

//###################### Macros ######################
#define uniq0(LOW0,HEIGHT0)	((HEIGHT0 << 8)|LOW0)		// 2x 8Bit 	--> 16Bit

#define uniq1(LOW1,HEIGHT1)	((HEIGHT1 << 8)|LOW1)		// 2x 8Bit 	--> 16Bit

#define uniq2(LOW2,HEIGHT2)	((HEIGHT2 << 8)|LOW2)		// 2x 8Bit 	--> 16Bit

#define uniq3(LOW3,HEIGHT3)	((HEIGHT3 << 8)|LOW3)		// 2x 8Bit 	--> 16Bit

#define sbi(ADDRESS,BIT) 	((ADDRESS) |= (1<<(BIT)))	// set Bit
#define cbi(ADDRESS,BIT) 	((ADDRESS) &= ~(1<<(BIT)))	// clear Bit
#define	toggle(ADDRESS,BIT)	((ADDRESS) ^= (1<<BIT))		// Bit umschalten

#define	bis(ADDRESS,BIT)	(ADDRESS & (1<<BIT))		// bit is set?
#define	bic(ADDRESS,BIT)	(!(ADDRESS & (1<<BIT)))		// bit is clear?

#define RF_BAUDRATE		2400		// Baudrate des RFM12

unsigned char data[128];

int32_t low0, hight0, low1, hight1, low2, hight2, low3, hight3;

unsigned char count;

int16_t temp0 = 0;  // Aussensensor Temperatur
int16_t temp1 = 0;  // Aussensensor Feuchtigkeit
int16_t temp2 = 0;  // Aussensensor Taupunkt
int16_t temp3 = 0;  // Aussensensor Batteriespannung
 
int main(void)
{

 char buf[20];

 lcd_init(LCD_DISP_ON);
 lcd_clrscr();
		
 sei(); //interrupts freigeben
 
 rf12_init();											// RF12 + IO Ports initialisieren
 rf12_config(RF_BAUDRATE, RF12FREQ(868.80), 0, NORMAL);	// Baudrate, Frequenz, Leistung (0=max, 7=min), Umgebungsbedingungen (QUIET, NORMAL, NOISY)
 rf12_rxmode();											// Empfang starten
	
    while (1) 
    {
		
		if (rx_data_in_buffer())     // RFM12-Daten empfangen?
		{							
			unsigned char i;
			
			count=rx_data(data);
			
			for (i=0; i<count; i++) // RFM12-Daten schreiben
			{						
		     low0   = data[0];
		     hight0 = data[1];
		     low1   = data[2];
		     hight1 = data[3];
		     low2   = data[4];
		     hight2 = data[5];
		     low3   = data[6];
		     hight3 = data[7];
			}
			
		}
		
	    temp0 = uniq0(low0,hight0);		// 2x 8Bit  --> 16Bit
	    temp1 = uniq1(low1,hight1);		// 2x 8Bit  --> 16Bit
	    temp2 = uniq2(low2,hight2);		// 2x 8Bit  --> 16Bit
	    temp3 = uniq3(low3,hight3);		// 2x 8Bit  --> 16Bit
		
		lcd_clrscr();
		
		lcd_gotoxy(0,0);
        sprintf( buf, "%i", temp0 );
		lcd_puts(buf);
		
		lcd_gotoxy(10,0);
        sprintf( buf, "%i", temp1 );
		lcd_puts(buf);
		
		lcd_gotoxy(0,1);
        sprintf( buf, "%i", temp2 );
		lcd_puts(buf);
		
		lcd_gotoxy(10,1);
        sprintf( buf, "%i", temp3 );
		lcd_puts(buf);
	 
    }	
	
	return 0;
}

