/****************************************************************************
 *             Sender Wetter-Funkbruecke
 *             =========================
 *					Uwe Berger; 2013
 *
 * Hardware
 * ========
 * -Standard ATMega8 (AVcc an Vcc; F_CPU (MHz)
 * -RFM12-Modul: Schaltung siehe http://www.mikrocontroller.net/topic/71682
 * -Hardware-TWI: siehe entsprechendes Atmel-Datenblatt zur MCU
 * -diverse Sensoren...
 *
 * Software
 * ========
 *
 * TWI-Master-Lib:
 * ---------------
 * -Peter Fleury (http://homepage.hispeed.ch/peterfleury/avr-software.html#libs)
 * 
 * 
 * RFM12-Lib:
 * ----------
 * -Benedikt K. (http://www.mikrocontroller.net/topic/71682)
 * 
 * 
 * SHT11-LIB
 * ---------
 * -A.K./Thorsten S. (http://www.mikrocontroller.net/topic/145736)
 * 
 * 
 * allgemeiner Programmablauf:
 * ---------------------------
 * -es werden zyklisch die verschiedensten angeschlossenen Sensoren 
 *  abgefragt und die ermittelten Werte in der Struktur msg abgelegt
 * -Messung von Vcc mit Vbg als Referenz; dieser Wert landet ebenfalls 
 *  in msg
 *  (http://www.mikrocontroller.net/articles/Pollin_Funk-AVR-Evaluationsboard)
 * -der Inhalt von msg wird zyklisch via Funk gesendet
 * -in den "Pausen" wird die MCU und das RFM12-Modul in den Sleep-Modus 
 *  geschickt, um Strom zu sparen; das Wecken des Systems erfolgt ueber 
 *  den WakeUp-Timer des RFM12-Modul 
 *  (http://www.mikrocontroller.net/topic/241964)
 * 
 * 
 * ---------
 * Have fun! 
 *
 ***************************************************************************/

#include "config.h"	       //IMPORTANT: it contains F_CPU definition, so include it before delay.h

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/io.h>
#include <stdint.h> 	    //definiert den Datentyp uint8_t
#include <stdlib.h>         //ntig fr Zahlumwandlung mit itoa
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "rfm12/rf12.h"

#define DEBUG_LED		0			// LED leuchtet waehrend Messen/Senden

#define RF_BAUDRATE		2400		// Baudrate des RFM12

#define RFM12_IRQ_DDR	DDRD		// INT0-Pin fuer RFM12-WakeUp
#define RFM12_IRQ_PORT	PORTD
#define RFM12_IRQ		(1<<PD2)

#define ADC_DUMMY_READS	10			// ein paar ADC-Parameter...
#define ADC_HOT_READS	20
#define V_BG			1100UL		// interner BandGap 1,1V

// Funkmessage...
struct msg_t {					// Offset
	uint16_t temperatur;	    // --> 0
	int16_t  feuchtigkeit;		// --> 2
	uint16_t taupunkt;			// --> 4
	uint16_t vcc;	            // --> 6
} msg;

uint16_t adc_read(uint8_t val_admux)
{
	uint32_t adcval;
	uint8_t i;
	ADMUX = val_admux;
	//
	//	ADC optimal zwischen 50 kHz und 200 kHz...
    //   F_CPU/200000 = min. Vorteiler
    //   F_CPU/50000  = max. Vorteiler
	// Vorteiler (64) und ADC einschalten
	ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS2);
	// sicherheitshalber ein paar Dummy-Messungen
	for(i=0; i < ADC_DUMMY_READS; i++) {
		ADCSRA |= (1<<ADSC);
    	while (ADCSRA&(1<<ADSC)) {}
		adcval = ADCW;
	}
	// reale Messungen
	adcval = 0;
	for(i=0; i < ADC_HOT_READS; i++) {
		ADCSRA |= (1<<ADSC);
		while (ADCSRA&(1<<ADSC)) {}
		adcval += ADCW;
	}    
	// ADC ausschalten
    ADCSRA &= ~(1<<ADEN);
    // Mittelwert zurueckgeben
	return adcval/ADC_HOT_READS;
}

ISR(INT0_vect) 
{
/*
#if DEBUG_LED
	// LED toggeln
	LED_PORT ^= (1 << LED_PIN);
#endif
*/
}

void rf12_sleep_with_wakeup(void) 
{
	//	
	// Wake-Up Timer 
	// 1 1 1 r4 r3 r2 r1 r0 m7 m6 m5 m4 m3 m2 m1 m0
	// Twake-up = M * 2R [ms]
	//
// 	rf12_trans(0xE730);  // ca. 6s
// 	rf12_trans(0xE6EB);  // ca. 15s
//	rf12_trans(0xE8EA);  // ca. 60s
  	rf12_trans(0xE9B0);  // ca. 150s
	//
	// Power Management
	// 1 0 0 0 0 0 1 0 er ebb et es ex eb ew dc
	// er  Enables the whole receiver chain (RF front end, baseband, 
	//     synthesizer, oscillator)
	// ebb The receiver baseband circuit can be separately switched on 
	//     (Baseband)
	// et  Switches on the PLL, the power amplifier, and starts the 
	//     transmission (If TX register is enabled) (Power amplifier, 
	//     synthesizer, oscillator)
	// es  Turns on the synthesizer (Synthesizer)
	// ex  Turns on the crystal oscillator (Crystal oscillator)
	// eb  Enables the low battery detector (Low battery detector)
	// ew  Enables the wake-up timer (Wake-up timer)
	// dc  Disables the clock output (pin 8) (Clock output buffer)
	//

	rf12_trans(0x8203);
	rf12_trans(0x0000);    // Interruptanforderungen lschen
}

//*************************************************
void radio_init(void) 
{
	rf12_init();										
	rf12_config(RF_BAUDRATE, RF12FREQ(868.80), 0, NORMAL);  // Baudrate, Frequenz, Leistung (0=max, 7=min), Umgebungsbedingungen (QUIET, NORMAL, NOISY)
	rf12_rxmode();		
}

int main(void)
{
 DDRD |= (1 << PD1) | (1 << PD0);       // damit ist dann PD ein Ausgang
 PORTD |= (1 << PD1);  // grne LED aus
 PORTD |= (1 << PD0);  // rote LED aus
 
 uint32_t val32;
 
 RFM12_IRQ_PORT &= ~RFM12_IRQ;
 EICRA |= (0<<ISC01) | (0<<ISC00);

 sei();

// Kurzes Aufleuchten beim Start
 PORTD &= ~(1 << PD0); // rote LED ein
 _delay_ms(200);
 PORTD |= (1 << PD0);  // rote LED aus

    while (1) 
    {
	
        #if DEBUG_LED
		PORTD &= ~(1 << PD0); // rote LED ein
	    #endif
	
		// RFM12-Modul initialisieren...
		radio_init();
		
		// interne Empfangsroutine, muss periodisch aufgerufen werden
		check_rx_packet();	
	
		// Vcc ermitteln (Ermittlung indirekt via Bandgab-Referenz)
		val32 = adc_read((1<<REFS0)|(1<<MUX3)|(1<<MUX2)|(1<<MUX1)|(0<<MUX0));
		msg.vcc = (V_BG*1024)/val32;
		msg.vcc = msg.vcc - 34;     // Korrektur
	
		msg.temperatur       = 123;  // Testwert
		msg.feuchtigkeit     = 456;  // Testwert
		msg.taupunkt         = 789;  // Testwert
		
		#if DEBUG_LED
	     // Sendekontrolle
         PORTD &= ~(1 << PD1); // grne LED ein
         _delay_ms(50);
         PORTD |= (1 << PD1);  // grne LED ein
         _delay_ms(50);
         PORTD &= ~(1 << PD1); // grne LED ein
         _delay_ms(50);
         PORTD |= (1 << PD1);  // grne LED ein
         _delay_ms(50);
         PORTD &= ~(1 << PD1); // grne LED ein
         _delay_ms(50);
         PORTD |= (1 << PD1);  // grne LED ein
         _delay_ms(50);
         PORTD &= ~(1 << PD1); // grne LED ein
         _delay_ms(50);
         PORTD |= (1 << PD1);  // grne LED ein
        #endif

		// Message via RFM12 versenden
		tx_data((unsigned char *)&msg, sizeof(msg));

        #if DEBUG_LED
        PORTD |= (1 << PD0);  // rote LED aus
		#endif
		
		// ...und alles schlafen schicken 
		// (Aufwachen via RFM12 --> INT0)
		rf12_sleep_with_wakeup();
		EIMSK |= (1 << INT0);
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		sleep_mode();                  
		EIMSK &= ~(1 << INT0); 

    }	
	
	return 0;
}

